#include "tables.cl"

#define INDX(i, j, k) ((i)*n.y*n.z+(j)*n.z+(k))
#define K(indx) ((indx)%n.z)
#define J(indx) (((indx)/n.z)%n.y)
#define I(indx) ((indx)/(n.z*n.y))

//#define MP(v) ((v.x)*n.y*n.z+(v.y)*n.z+(v.z))
//#define IJK(mp) ((int3)(((mp)%n.z), (((mp)/n.z)%n.y), ((mp)/(n.z*n.y))))
//#define EDGE_DELTA_IJK (indx) ((int3)( ((indx)%2), (((indx)/2)%2), ((indx)/4)))

#define EDGE_ID(i, j, k) ((i)*4+(j)*2+(k))
#define EDGE_DK(indx) ((indx)%2)
#define EDGE_DJ(indx) (((indx)/2)%2)
#define EDGE_DI(indx) ((indx)/4)

inline float getX(const int mp, const int3 n, const float3 o, 
                  const float3 e1, const float3 e2, const float3 e3)
{
    float3 ijk = (float3)(I(mp), J(mp), K(mp));
    float3 ex = (float3)(e1.x, e2.x, e3.x);
    return o.x + dot(ijk, ex);
}

inline float getY(const int mp, const int3 n, const float3 o, 
                  const float3 e1, const float3 e2, const float3 e3)
{
    float3 ijk = (float3)(I(mp), J(mp), K(mp));
    float3 ey = (float3)(e1.y, e2.y, e3.y);
    return o.y + dot(ijk, ey);
}

inline float getZ(const int mp, const int3 n, const float3 o, 
                  const float3 e1, const float3 e2, const float3 e3)
{
    float3 ijk = (float3)(I(mp), J(mp), K(mp));
    float3 ez = (float3)(e1.z, e2.z, e3.z);
    return o.z + dot(ijk, ez);
}

inline float3 getXYZ(
    const int mp, 
    const int3 n, 
    const float3 o, 
    const float3 e1, 
    const float3 e2, 
    const float3 e3)
{
    float3 ijk = (float3)(I(mp), J(mp), K(mp));
    float3 ex = (float3)(e1.x, e2.x, e3.x);
    float3 ey = (float3)(e1.y, e2.y, e3.y);
    float3 ez = (float3)(e1.z, e2.z, e3.z);
    return o+(float3)(dot(ijk, ex), dot(ijk, ey), dot(ijk, ez));
}


__kernel void fillMpIndex(
    __global float* field,
    __global uchar* mpIndex,
    const float isoValue, 
    const int3 n)
{
    int mp = get_global_id(0);
    int totalMPs = n.x*n.y*n.z;
    if(mp<totalMPs)
    {
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);
        bool firstBit = field[mp]<isoValue;
        bool zlimit = (k+1<n.z);
        bool ylimit = (j+1<n.y);
        bool xlimit = (i+1<n.x);
        mpIndex[mp] = firstBit;
        //If vertex is outside the boundary then fill its bit with the value of
        //first vertex. This is useful for extracting edges later.
        
        mpIndex[mp] += 2   * ((zlimit && (field[INDX(i, j, k+1)]<isoValue)) || (!zlimit && firstBit));
        mpIndex[mp] += 4   * ((ylimit && (field[INDX(i, j+1, k)]<isoValue)) || (!ylimit && firstBit) );
        mpIndex[mp] += 8   * ((ylimit && zlimit && (field[INDX(i, j+1, k+1)]<isoValue)) || (!(zlimit&&ylimit) && firstBit));
        mpIndex[mp] += 16  * ((xlimit && (field[INDX(i+1, j, k)]<isoValue)) || (!xlimit && firstBit) );
        mpIndex[mp] += 32  * ((xlimit && zlimit && (field[INDX(i+1, j, k+1)]<isoValue)) || (!(xlimit&&zlimit) && firstBit) );
        mpIndex[mp] += 64  * ((xlimit && ylimit && (field[INDX(i+1, j+1, k)]<isoValue)) || (!(xlimit&&ylimit) && firstBit));  
        mpIndex[mp] += 128 * ((xlimit && ylimit && zlimit && (field[INDX(i+1, j+1, k+1)]<isoValue)) || (!(xlimit&&ylimit&&zlimit) && firstBit));                                              
    }

}

inline uchar mpIndex2edgeIndex(uchar mpIndex)
{
    uchar edgeIndex = 0;
    uchar firstBit = mpIndex&1;

    for(uchar i=1;i<8;i++)
    {
        uchar mask = 1 << i;
        edgeIndex += mask * (firstBit ^  ((mpIndex&mask)>0));

    }
    return edgeIndex;
}

inline int vertsNum(uchar mpIndex)
{
    int vertsNum = 0;
    uchar firstBit = mpIndex&1;

    for(uchar i=1;i<8;i++)
    {
        uchar mask = 1 << i;
        vertsNum += (firstBit ^  ((mpIndex&mask)>0));

    }
    return vertsNum;
}

__kernel void enumActiveMPs(__global uchar* mpIndex, 
                            __global int* mpActiveId,
                            __global int* activeMPs, const int totalMPs)
{
    int mp = get_global_id(0);
    if(mp<totalMPs)
    {
        if(mpIndex2edgeIndex(mpIndex[mp])>0)
        {
            activeMPs[mpActiveId[mp]]= mp;
        }
    }
    else if(mp == totalMPs)
    {
        activeMPs[mpActiveId[totalMPs-1]]= totalMPs-1;
    }
}

__kernel void calcVoxNormals(__global float* field, __global float* voxNormals, 
                             const int3 n, const float3 o, 
                             const float3 e1, const float3 e2, const float3 e3)
{
    int mp = get_global_id(0);
    int totalMPs = n.x*n.y*n.z;
    if(mp<totalMPs)
    {
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);

        float dfx,dfy,dfz, dx,dy,dz;
        
        
        int mp1,mp2;
        
        mp1 = INDX(i-1, j, k);
        mp2 = INDX(i+1, j, k);            

            
//        float3 xyz1 = getXYZ(mp1, n, o, e1, e2, e3);
//        float3 xyz2 = getXYZ(mp2, n, o, e1, e2, e3);

//        float3 delta = xyz2 - xyz1;
//        float df = field[mp2] - field[mp1];        
//        voxNormals[mp] = df/delta;
        
        if(i-1<0)
            mp1 = mp;
        if(i+1>=n.x)
            mp2 = mp;        
        dfx = field[mp2] - field[mp1];
        dx = getX(mp2, n, o, e1, e2, e3) - getX(mp1, n, o, e1, e2, e3);
        
        
        mp1 = INDX(i, j-1, k);
        mp2 = INDX(i, j+1, k);            
        if(j-1<0)
            mp1 = mp;
        if(j+1>=n.y)
            mp2 = mp;         
        dfy = field[mp2] - field[mp1];
        dy = getY(mp2, n, o, e1, e2, e3) - getY(mp1, n, o, e1, e2, e3);
        
        mp1 = INDX(i, j, k-1);
        mp2 = INDX(i, j, k+1);            
        if(k-1<0)
            mp1 = mp;
        if(k+1>=n.z)
            mp2 = mp;         
        dfz = field[mp2] - field[mp1];
        dz = getZ(mp2, n, o, e1, e2, e3) - getZ(mp1, n, o, e1, e2, e3);

        voxNormals[mp*3] = dfx/dx;
        voxNormals[mp*3+1] = dfy/dy;
        voxNormals[mp*3+2] = dfz/dz;
    }
}


__kernel void calcVertsAndNorms(__global float* verts, __global float* normals, 
                                const float isoValue, 
                                __global float* field, 
                                __global float* voxNormals, 
                                __global char* mpIndex, 
                                __global int* ampVertsBase,
                                __global int* activeMPs, const int totalActive,
                                const int3 n, const float3 o, const float3 e1, 
                                const float3 e2, const float3 e3)
{
    int amp = get_global_id(0);
    if(amp<totalActive)
    {
        int mp0 = activeMPs[amp];
        int i0 = I(mp0);
        int j0 = J(mp0);
        int k0 = K(mp0);
        
        float f0 = field[mp0];
        float3 xyz0 = getXYZ(mp0, n, o, e1, e2, e3);        
        float3 n0 = (float3) (voxNormals[mp0*3], voxNormals[mp0*3+1], voxNormals[mp0*3+2]);

        uchar mp0Index = mpIndex[mp0];
        uchar edgeIndex = 0;
        uchar vertsNum = 0;
        
        uchar firstBit = mp0Index&1;

        for(uchar i=1;i<8;i++)
        {
            uchar mask = 1 << i;
            uchar tmp =  (firstBit ^  ((mp0Index&mask)>0));
            edgeIndex += mask * tmp;
            vertsNum += tmp;
            
        }
        
        for(uchar v=0;v<vertsNum;v++)
        {
            int edgeId = vert2EdgeId[v][edgeIndex/2];
            int i1 = i0+EDGE_DI(edgeId);
            int j1 = j0+EDGE_DJ(edgeId);
            int k1 = k0+EDGE_DK(edgeId);

            int mp1 = INDX(i1,j1,k1);
            
            float f1 = field[mp1];
            float3 xyz1 = getXYZ(mp1, n, o, e1, e2, e3);        
            float3 n1 = (float3) (voxNormals[mp1*3], voxNormals[mp1*3+1], voxNormals[mp1*3+2]);
            
            float alpha = (isoValue - f0) / (f1 - f0);

            int vAddress = ampVertsBase[amp]+v;
            float3 xyz = mix(xyz0, xyz1, alpha);
            vstore3(xyz, vAddress, verts);
            
            float3 n = mix(n0, n1, alpha);
            n = normalize(n);
            vstore3(n, vAddress, normals);
        }    
    }
}

__kernel void calcTriangs(__global float* field, __global int* triangs, 
                          __global int* voxTriangsBase, 
                          __global uchar* mpIndex, __global int* ampVertsBase,
                          __global int* activeMPs, __global int* mpActiveId,
                          const int totalActive, 
                          const float isoValue, const int3 n)
{
    int amp = get_global_id(0);
    if(amp<totalActive)
    {
        int mp = activeMPs[amp];
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);
        if(!(i<n.x-1&&j<n.y-1&&k<n.z-1))
        {
            return;
        }
        uchar myIndex = mpIndex[mp];
        int triangleShift = 0;
        for(int tetra=0;tetra<6;tetra++)
        {
            int tetraIndx = 0;
            int bit = 1;
            for(int tmp=0;tmp<4;tmp++)
            {
                uchar mask = 1 << tetraMP2voxMP[tetra][tmp];
                tetraIndx+= ((myIndex&mask)>0)*bit;
                bit*=2;
            }
            int tetraVerts[6];
            int v = 0;
            for(int tmp1=0;tmp1<4;tmp1++)
            {
                for(int tmp2=tmp1+1;tmp2<4;tmp2++)
                {
                    int min, max;
                    if(tetraMP2voxMP[tetra][tmp1]>tetraMP2voxMP[tetra][tmp2])
                    {
                        min = tetraMP2voxMP[tetra][tmp2];
                        max = tetraMP2voxMP[tetra][tmp1];
                    }
                    else
                    {
                        max = tetraMP2voxMP[tetra][tmp2];
                        min = tetraMP2voxMP[tetra][tmp1];
                    }
                    int associatedMP = INDX(i+EDGE_DI(min), j+EDGE_DJ(min), k+EDGE_DK(min));
                    int base = ampVertsBase[mpActiveId[associatedMP]];
                    
                    
                    int dI = EDGE_DI(max)-EDGE_DI(min);
                    int dJ = EDGE_DJ(max)-EDGE_DJ(min);
                    int dK = EDGE_DK(max)-EDGE_DK(min);
                    
                    int edgeId = EDGE_ID(dI,dJ,dK);
                    
                    
                    int shift = 
                        edge2VertShift[edgeId-1][mpIndex2edgeIndex(mpIndex[associatedMP])/2];
                    if(shift>=0)
                        tetraVerts[v] = base + shift;
                    else
                        tetraVerts[v] = -1;
                    v++;
                }
            }
            
            for(int triangle=0;triangle<2;triangle++)
            {
                if(tetraTriangVerts[tetraIndx][3*triangle] < 0)
                        break;
                for(int vert = 0; vert < 3; vert++)
                {
                    int triangAddress = voxTriangsBase[amp]+triangleShift;

                    triangs[triangAddress*3+vert] =  
                    tetraVerts[tetraTriangVerts[tetraIndx][3*triangle+vert]];
                }
                triangleShift++;
            }            
        }
    }
}
