#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <assert.h>


#if defined(WIN32)
	#define GLEW_STATIC
	#include <GL/glew.h>
	#include <windows.h>
#else
	#include <GL/glew.h>
#endif


#if defined(__APPLE__) || defined(MACOSX)
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <GLUT/glut.h>
#else
	#include <GL/glut.h>
	#include <GL/gl.h>
	#include <GL/glu.h>
#endif

#include <math.h>
#include <string.h>

#include "oclTetra.h"

int vertsNum = 0;
int triangsNum = 0;

int vertsBuff, normsBuff, triangsBuff;
float isoValue;

void generateSurface()
{
        printf("Generating isosurface for isovalue %g ...\n", isoValue);
        glDeleteBuffers(1, &vertsBuff);
        glDeleteBuffers(1, &normsBuff);
        glDeleteBuffers(1, &triangsBuff);
        oclTetraGenSurf(isoValue, &vertsNum, &triangsNum, 
                        &vertsBuff, &normsBuff, &triangsBuff);
        printf("Isosurfave is generated with %d vertices and %d triangles ...\n", vertsNum, triangsNum);
}

float pi;
float r, theta, phi;
int smooth = 1;
void init(void)
{
    GLfloat mat_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };    
    GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_shininess[] = { 100.0 };
    GLfloat white_light[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat lmodel_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
    GLfloat light_position[] = { 0.0, 0.0, 1.0, 0.0 };    
    
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glShadeModel (GL_SMOOTH);

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);    
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
    
    glMaterialfv(GL_BACK, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_BACK, GL_DIFFUSE, mat_diffuse);    
    glMaterialfv(GL_BACK, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_BACK, GL_SHININESS, mat_shininess);    

    glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
    glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);                    
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable (GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

}

void drawGrid(float spacing, float min, float max)
{
    glDisable(GL_LIGHTING);
    float coord;
    glLineWidth(1);
    glBegin (GL_LINES);
    //Draw X-Y Plane            
    glColor3f(0.2, 0.2, 0.2);
    coord = spacing;
    while (coord <= max)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord += spacing;
    }
    coord = -spacing;
    while (coord >= min)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord -= spacing;
    }
    glEnd();
    glLineWidth(3);
    glBegin(GL_LINES);
    //Draw Axes    

    glColor3f(0.3, 0, 0);
    glVertex3f(min, 0, 0);
    glVertex3f(0, 0, 0);
    glColor3f(1, 0, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(max, 0, 0);

    glColor3f(0, 0.3, 0);
    glVertex3f(0, min, 0);
    glVertex3f(0, 0, 0);
    glColor3f(0, 1, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, max, 0);

    glColor3f(0, 0, 0.3);
    glVertex3f(0, 0, min);
    glVertex3f(0, 0, 0);
    glColor3f(0, 0, 1);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, max);

    glEnd();
    glEnable(GL_LIGHTING);
}


void display(void)
{
    //generateSurface();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //
    glLoadIdentity();
    float x = r * sin(theta) * cos(phi);
    float y = r * sin(theta) * sin(phi);
    float z = r * cos(theta);

    gluLookAt(x, y, z, 0.0, 0.0, 0.0, 0.0f, 0.0f, 1.0f);

    //Draw grid

    drawGrid(1, -30, 30);              
    
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    
    glBindBuffer(GL_ARRAY_BUFFER, vertsBuff);
    glVertexPointer(3, GL_FLOAT, 0, 0);    

    glBindBuffer(GL_ARRAY_BUFFER, normsBuff);
    glNormalPointer(GL_FLOAT, 0, 0);
    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, triangsBuff);
            
    
    glDrawElements(GL_TRIANGLES, triangsNum*3, GL_UNSIGNED_INT, 0);              
    
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);      
    
    glutSwapBuffers();
}

int gw, gh;
void reshape(int w, int h)
{
    // Prevent a divide by zero, when window is too short
    // (you cant make a window of zero width).
    if (h == 0)
        h = 1;

    float ratio = 1.0 * w / h;

    // Reset the coordinate system before modifying
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();

    // Set the viewport to be the entire window
    glViewport(0, 0, w, h);

    // Set the correct perspective.
    gluPerspective(45, ratio, 0.5, 500);
    glMatrixMode (GL_MODELVIEW);
    gh = h;
    gw = w;
}

int fullscreen = 0;
int screen_pos_x;
int screen_pos_y;
int screen_w;
int screen_h;


void processNormalKeys(unsigned char key, int x, int y)
{
    //Navigation
    if (key == 'w' || key == 'W')
    {
        if (theta - 0.1 > 0)
            theta -= 0.1;
    }
    if (key == 's' || key == 'S')
    {
        if (theta + 0.1 < pi)
            theta += 0.1;
    }

    if (key == 'a' || key == 'A')
    {
        phi -= 0.1;
    }
    if (key == 'd' || key == 'D')
    {
        phi += 0.1;
    }

    if (key == 'q' || key == 'Q')
    {
        r = r * 1.2;
    }

    if (key == 'e' || key == 'E')
    {
        r = r / 1.2;
    }

    if (key == 'p' || key == 'P')
    {
        printf("Camera Position:");
        printf("r = %lg, theta = %lg, phi= %lg ", r, theta, phi);
        float x = r * sin(theta) * cos(phi);
        float y = r * sin(theta) * sin(phi);
        float z = r * cos(theta);
        printf("(x = %lg, y = %lg, z= %lg).\n", x, y, z);
    }

    if (key == 'f' || key == 'F')
    {
        fullscreen = !fullscreen;
        if (fullscreen)
        {
            screen_pos_x = glutGet((GLenum) GLUT_WINDOW_X);
            screen_pos_y = glutGet((GLenum) GLUT_WINDOW_Y);
            screen_w = gw;
            screen_h = gh;
            glutFullScreen();
        }
        else
        {
            glutReshapeWindow(screen_w, screen_h);
            glutPositionWindow(screen_pos_x, screen_pos_y);
        }
    }
    if (key == 'i' || key == 'I')
    {
        printf("Isovlaue: %lg\n", isoValue);
    }
    if (key == 'u' || key == 'U')
    {
        if(smooth)
        {
            glShadeModel (GL_FLAT);
            smooth = 0;
        }
        else
        {
            glShadeModel (GL_SMOOTH);
            smooth = 1;
        } 
    }
    if (key == 'z' || key == 'Z')
    {        
        isoValue /= 1.1;
        
        generateSurface();
    }
    if (key == 'x' || key == 'X')
    {
        isoValue *= 1.1;
        generateSurface();
    }
    if (key == 27)
    {
        oclTetraClose();
        exit(0);
    }
}


float mouse_x = 0;
float mouse_y = 0;
float old_theta = 0, old_phi = 0;
int pressed = 0;

void mouseButton(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON)
    {

        // when the button is released
        if (state == GLUT_UP)
        {
            pressed = 0;
        }
        else
        { // state = GLUT_DOWN
            mouse_x = x;
            mouse_y = y;
            old_theta = theta;
            old_phi = phi;
            pressed = 1;
        }
    }
    // Wheel reports as button 3(scroll up) and button 4(scroll down)
    if ((button == 3 && state == GLUT_DOWN)
            || (button == 4 && state == GLUT_DOWN))
    {
        if (button == 4)
            r *= 1.2;
        else
            r /= 1.2;
    }
}

void mouseMove(int x, int y)
{
    if (pressed)
    {
        theta = old_theta + (float) (mouse_y - y) / gh * pi;
        if (theta < 0)
            theta = 0.001;
        if (theta > pi)
            theta = pi - 0.001;
        phi = old_phi + (float) (mouse_x - x) / gw * pi;
    }
}

char *textFileRead(char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

int main(int argc, char** argv)
{

    if(argc<3)
    {
        fprintf(stderr, 
                "Please specify field datafile followed by isovalue.\n");
        exit(1);
    }
    
    char* fileName = argv[1];
    isoValue = atof(argv[2]);
  
    
    pi = atan(1)*4;
    r = 60;
    theta = pi / 2.5;
    phi = pi / 4.0;
    

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(800, 600);
    glutInitWindowPosition(0, 0);
    glutCreateWindow("Marching Tetrahedra Using OpenCL");

    glutDisplayFunc(display);
    glutIdleFunc(display);    
    glutReshapeFunc(reshape);
    glutKeyboardFunc(processNormalKeys);
    glutMouseFunc(mouseButton);
    glutMotionFunc(mouseMove);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0"))
    {
        printf("OpenGL 2.0 not supported\n");
        exit(1);
    }

    init();
    
    char* control = textFileRead("control.txt");
    printf("%s", control);
    free(control);

    
    oclTetraInit(fileName);        
    
    generateSurface(); 
           
    glutMainLoop();
    
    return 0;
}


