#include <CL/cl.h>

//Defining oclScanner as an opaque object. User can only declare pointers to oclScanner
typedef struct opaque oclScanner;


//headerFilename is an opencl kerenel file that should have the following format:
//
//#define IN_TYPE       int         //Datatype of input array
//#define OUT_TYPE      int         //Datatype of output array
//#define BIN_OP(A,B)   ((A)+(B))   //Binary operation used for scan
//#define IDENTITY      0           //The Identity of the binary operation
//#define PARA_TYPE     const int   //Datatype of the parameter used in preprocess
//
// Preprocess input array before performing scan
//OUT_TYPE Preprocess(IN_TYPE in, PARA_TYPE para)
//{
//    OUT_TYPE out;
//    
//    out = in;
//    
//    return out;
//}
//

oclScanner* oclCreateScanner(
    cl_context context, 
    cl_command_queue queue, 
    cl_device_id device, 
    unsigned int workgroupSize, 
    unsigned int maxWorkItemsCount, 
    const char* headerFilename, 
    size_t outSize, 
    int paraCount,
    size_t* paraSizes,
    int* err);

void oclScan( 
    oclScanner* scanner, 
    cl_mem output_data,  
    cl_mem input_data,
     unsigned int count, 
     void** paras);   
    
void oclReleaseScanner(oclScanner* scanner);
