#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>


#if defined(WIN32)
	#define GLEW_STATIC
	#include <GL/glew.h>
	#include <windows.h>
	
#elif defined(__APPLE__) || defined(MACOSX)
    #include <OpenCL/opencl.h>
	#include <GL/glew.h>
	
    #include <OpenCL/cl_gl_ext.h>
#else
    #include <GL/glew.h>
    #include <GL/glx.h>
    
    #include <CL/cl.h>
    #include <CL/cl_gl.h>
#endif

#define KER_DIR "./kernels/"

#include "oclScan.h"
#include "oclTetra.h"

cl_platform_id platform;
cl_context context;
cl_command_queue queue;
cl_device_id device;
cl_program program;
cl_int error = 0;

//To do: Currently the code doesn't check the maximum number of work items in each dimension.
unsigned int workgroupSize = 256;
unsigned int maxWorkItemsCount = 256;

oclScanner* scanner, *activeScanner, *vertsScanner, *triangsScanner;
size_t vertsScanParaSizes [] = {sizeof(cl_mem)};
size_t triangsScanParaSizes [] = {sizeof(cl_mem), sizeof(cl_int3)};
void* vertsScanParas [] = {0};
void* triangsScanParas [] = {0, 0};
#define GLOBAL_WORK_SIZE(ACTUAL_WORK_SIZE) (((((ACTUAL_WORK_SIZE)-1)/workgroupSize)+1)*workgroupSize)
#define min(A,B) ((A) < (B) ? (A) : (B))

cl_kernel fillMpIndex_kernel,
          enumActiveMPs_kernel, 
          calcVoxNormals_kernel, 
          calcVertsAndNorms_kernel, 
          calcTriangs_kernel;

cl_int3 count;
cl_float3 origin, 
          e1, 
          e2, 
          e3;
          
int totalMPs, 
    totalActive, 
    totalVerts, 
    totalTriangs;
    
float isoValue;

cl_mem field, 
       voxNormals,
       mpIndex,
       mpActiveId,
       activeMPs, 
       ampVertsBase,

       verts, 
       normals, 
       voxTriangsBase, 
       triangs;


#define INDX(i, j, k) ((i)*count.s[1]*count.s[2]+(j)*count.s[2]+(k))
void loadField(char* fileName)
{
    printf("Loading field data from file: '%s' ...\n", fileName);
    FILE* fp = fopen(fileName, "r");   
    if(fp==0)
    {
        fprintf(stderr, "Can not open file!\n");
        exit(1);
    }

    char c = '0';
    while(c!='\n' && c!=EOF) c = fgetc(fp);
    c = '0';
    while(c!='\n' && c!=EOF) c = fgetc(fp);
    
    fscanf(fp, " %*d %g %g %g\n", &origin.s[0], &origin.s[1] ,&origin.s[2]);
    
    fscanf(fp, " %d %g %g %g\n", &count.s[0], &e1.s[0], &e1.s[1], &e1.s[2]);
    fscanf(fp, " %d %g %g %g\n", &count.s[1], &e2.s[0], &e2.s[1], &e2.s[2]);
    fscanf(fp, " %d %g %g %g\n", &count.s[2], &e3.s[0], &e3.s[1], &e3.s[2]);
    totalMPs = count.s[0] * count.s[1] * count.s[2];
    float* field_temp = (float*) malloc(sizeof(float) * totalMPs);
    for(int i=0;i<count.s[0];i++)
    {   
        for(int j=0;j<count.s[1];j++)
        {
            for (int k=0;k<count.s[2];k++)
            {
                fscanf(fp, " %g ", &field_temp[INDX(i, j, k)]);
            }
        }
    }    
    fclose(fp);
    
    // Allocate memory & transfer to GPU
    field = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, sizeof(float)*totalMPs, field_temp, &error);
    assert (error == CL_SUCCESS);
    
    free(field_temp);      
}



void calcVoxNormals()
{
    // Allocate memory
    voxNormals = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)*3*totalMPs, 0, &error);
    assert (error == CL_SUCCESS); 
    
    // Enqueuing parameters
    error = clSetKernelArg(calcVoxNormals_kernel, 0, sizeof(cl_mem), &field);
    error |= clSetKernelArg(calcVoxNormals_kernel, 1, sizeof(cl_mem), &voxNormals);
    error |= clSetKernelArg(calcVoxNormals_kernel, 2, sizeof(cl_int3), &count);
    error |= clSetKernelArg(calcVoxNormals_kernel, 3, sizeof(cl_float3), &origin);
    error |= clSetKernelArg(calcVoxNormals_kernel, 4, sizeof(cl_float3), &e1);
    error |= clSetKernelArg(calcVoxNormals_kernel, 5, sizeof(cl_float3), &e2);
    error |= clSetKernelArg(calcVoxNormals_kernel, 6, sizeof(cl_float3), &e3);         
    assert(error == CL_SUCCESS);

    // Launching kernel
    const size_t global_ws = GLOBAL_WORK_SIZE(totalMPs);
    error = clEnqueueNDRangeKernel(queue, calcVoxNormals_kernel, 1, NULL, &global_ws, &workgroupSize, 0, NULL, NULL);
    assert(error == CL_SUCCESS);

}

void fillMpIndex()
{
    // Enqueuing parameters
    error = clSetKernelArg(fillMpIndex_kernel, 0, sizeof(cl_mem), &field);
    error |= clSetKernelArg(fillMpIndex_kernel, 1, sizeof(cl_mem), &mpIndex);
    error |= clSetKernelArg(fillMpIndex_kernel, 2, sizeof(cl_float), &isoValue);
    error |= clSetKernelArg(fillMpIndex_kernel, 3, sizeof(cl_int3), &count);
    assert(error == CL_SUCCESS);

    // Launching kernel
    const size_t global_ws = GLOBAL_WORK_SIZE(totalMPs);
    error = clEnqueueNDRangeKernel(queue, fillMpIndex_kernel, 1, NULL, &global_ws, &workgroupSize, 0, NULL, NULL);
    assert(error == CL_SUCCESS);

}

void enumActiveMPs()
{    
    // Scan
    oclScan(activeScanner, mpActiveId, mpIndex, totalMPs, NULL);  
    
    //Warning: last mesh point will always be ignored
    clEnqueueReadBuffer(queue, mpActiveId, CL_TRUE,  sizeof(int)*(totalMPs-1), sizeof(int), &totalActive, 0, NULL, NULL);

    //printf("%d\n", totalActive);
    // Allocate memory
    //We have one additional element, so that prefix scan automatically calculates the total sum
    activeMPs = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(int)*(totalActive+1), 0, &error);
    assert (error == CL_SUCCESS);
    
    // Enqueuing parameters
    error = clSetKernelArg(enumActiveMPs_kernel, 0, sizeof(cl_mem), &mpIndex);
    error |= clSetKernelArg(enumActiveMPs_kernel, 1, sizeof(cl_mem), &mpActiveId);
    error |= clSetKernelArg(enumActiveMPs_kernel, 2, sizeof(cl_mem), &activeMPs);
    error |= clSetKernelArg(enumActiveMPs_kernel, 3, sizeof(cl_int), &totalMPs);
    assert(error == CL_SUCCESS);

    // Launching kernel
    const size_t global_ws = GLOBAL_WORK_SIZE(totalMPs+1);
    error = clEnqueueNDRangeKernel(queue, enumActiveMPs_kernel, 1, NULL, &global_ws, &workgroupSize, 0, NULL, NULL);
    assert(error == CL_SUCCESS);

}

void enumVerts()
{
    // Allocate memory
    ampVertsBase = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(int)*(totalActive+1), 0, &error);
    assert (error == CL_SUCCESS);        

    // Prefix scan
    vertsScanParas[0] = &mpIndex;
    oclScan(vertsScanner, ampVertsBase, activeMPs, totalActive+1, vertsScanParas); 
}


void calcVertsAndNorms(int* vertsBuff, int* normsBuff)
{

    clEnqueueReadBuffer(queue, ampVertsBase, CL_TRUE,  sizeof(int)*(totalActive), sizeof(int), &totalVerts, 0, NULL, NULL);

    // Allocate OpenGL buffers
    glGenBuffers(1, vertsBuff);
    glBindBuffer(GL_ARRAY_BUFFER, *vertsBuff);
    glBufferData(GL_ARRAY_BUFFER, sizeof(float)*3*totalVerts, NULL, GL_STATIC_DRAW);
     
    glGenBuffers(1, normsBuff);
    glBindBuffer(GL_ARRAY_BUFFER, *normsBuff);
    glBufferData(GL_ARRAY_BUFFER, sizeof(float)*3*totalVerts, NULL, GL_STATIC_DRAW);
    
    glFinish(); 
    
    // Attach OpenCL memory to OpenGL buffers
    verts = clCreateFromGLBuffer(context, CL_MEM_READ_WRITE, *vertsBuff, &error);
    assert (error == CL_SUCCESS); 
    clEnqueueAcquireGLObjects(queue, 1, &verts, 0, NULL, NULL);    
    
    normals = clCreateFromGLBuffer(context, CL_MEM_READ_WRITE, *normsBuff, &error);
    assert (error == CL_SUCCESS); 
    clEnqueueAcquireGLObjects(queue, 1, &normals, 0, NULL, NULL);      
    assert (error == CL_SUCCESS);


    // Enqueuing parameters
    error = clSetKernelArg(calcVertsAndNorms_kernel, 0, sizeof(cl_mem), &verts);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 1, sizeof(cl_mem), &normals);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 2, sizeof(cl_float), &isoValue);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 3, sizeof(cl_mem), &field);    
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 4, sizeof(cl_mem), &voxNormals);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 5, sizeof(cl_mem), &mpIndex);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 6, sizeof(cl_mem), &ampVertsBase);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 7, sizeof(cl_mem), &activeMPs); 
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 8, sizeof(cl_int), &totalActive);           
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 9, sizeof(cl_int3), &count);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 10, sizeof(cl_float3), &origin);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 11, sizeof(cl_float3), &e1);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 12, sizeof(cl_float3), &e2);
    error |= clSetKernelArg(calcVertsAndNorms_kernel, 13, sizeof(cl_float3), &e3);         
    assert(error == CL_SUCCESS);

    // Launching kernel
    const size_t global_ws = GLOBAL_WORK_SIZE(totalActive);
    error = clEnqueueNDRangeKernel(queue, calcVertsAndNorms_kernel, 1, NULL, &global_ws, &workgroupSize, 0, NULL, NULL);
    assert(error == CL_SUCCESS);

}


void enumTriangs()
{
    // Allocate memory
    voxTriangsBase = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(int)*(totalActive+1), 0, &error);
   
   // Prefix scan
   triangsScanParas[0] =  &mpIndex;
   triangsScanParas[1] =  &count;
   oclScan(triangsScanner, voxTriangsBase, activeMPs, totalActive+1,  triangsScanParas);


}

void calcTriangs(int* triangsBuff)
{
    clEnqueueReadBuffer(queue, voxTriangsBase, CL_TRUE,  sizeof(int)*(totalActive), sizeof(int), &totalTriangs, 0, NULL, NULL);


    // Allocate OpenGL buffers 
    glGenBuffers(1, triangsBuff);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, *triangsBuff);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(int)*3*totalTriangs, NULL, GL_STATIC_DRAW);
    
    glFinish();
    
    // Attach OpenCL memory to OpenGL buffers
    triangs = clCreateFromGLBuffer(context, CL_MEM_READ_WRITE, *triangsBuff, &error);
    assert (error == CL_SUCCESS); 
    error = clEnqueueAcquireGLObjects(queue, 1, &triangs, 0, NULL, NULL);
    assert (error == CL_SUCCESS); 
    
    // Enqueuing parameters
    error = clSetKernelArg(calcTriangs_kernel, 0, sizeof(cl_mem), &field);
    error |= clSetKernelArg(calcTriangs_kernel, 1, sizeof(cl_mem), &triangs);
    error |= clSetKernelArg(calcTriangs_kernel, 2, sizeof(cl_mem), &voxTriangsBase);
    error |= clSetKernelArg(calcTriangs_kernel, 3, sizeof(cl_mem), &mpIndex);        
    error |= clSetKernelArg(calcTriangs_kernel, 4, sizeof(cl_mem), &ampVertsBase);      
    error |= clSetKernelArg(calcTriangs_kernel, 5, sizeof(cl_mem), &activeMPs);
    error |= clSetKernelArg(calcTriangs_kernel, 6, sizeof(cl_mem), &mpActiveId);    
    error |= clSetKernelArg(calcTriangs_kernel, 7, sizeof(cl_int), &totalActive);       
    error |= clSetKernelArg(calcTriangs_kernel, 8, sizeof(cl_float), &isoValue);
    error |= clSetKernelArg(calcTriangs_kernel, 9, sizeof(cl_int3), &count);
    assert(error == CL_SUCCESS);

    // Launching kernel
    const size_t global_ws = GLOBAL_WORK_SIZE(totalActive);
    error = clEnqueueNDRangeKernel(queue, calcTriangs_kernel, 1, NULL, &global_ws, &workgroupSize, 0, NULL, NULL);
    assert(error == CL_SUCCESS);
}



void oclTetraGenSurf(float _isoValue, int* vertsNum, int* triangsNum, int* vertsBuff, int* normsBuff, int* triangsBuff)
{   
    isoValue = _isoValue;        

    
    //Generating Vertices & Normals

    
    fillMpIndex();
    enumActiveMPs();
    enumVerts();      
           
       
    calcVertsAndNorms(vertsBuff, normsBuff); 
    //Generating Triangles    
 
    enumTriangs();
    calcTriangs(triangsBuff);    
   
    //Release Memory
    clReleaseMemObject(ampVertsBase);
    clReleaseMemObject(activeMPs);

    clReleaseMemObject(voxTriangsBase);    

    error = clEnqueueReleaseGLObjects(queue, 1, &verts, 0, NULL, NULL);
    assert (error == CL_SUCCESS); 
    error = clEnqueueReleaseGLObjects(queue, 1, &normals, 0, NULL, NULL);
    assert (error == CL_SUCCESS); 
    error = clEnqueueReleaseGLObjects(queue, 1, &triangs, 0, NULL, NULL);
    assert (error == CL_SUCCESS); 

    clReleaseMemObject(triangs); 
    clReleaseMemObject(verts);
    clReleaseMemObject(normals);    
    
    //Prepare for OpenGL
    error = clFinish(queue);
    assert (error == CL_SUCCESS); 
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    //Write back the total numbers of vertices and triangles
    *vertsNum = totalVerts;
    *triangsNum = totalTriangs;
        
}

static char *loadTextFile(const char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

void oclTetraInit(char* fieldFileName)
{               
    //Platform
    error = clGetPlatformIDs(1, &platform, NULL);
    assert (error == CL_SUCCESS);
    
    //Device
    error = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    assert (error == CL_SUCCESS);
    
    // Check OpenCL-OpenGL interop
    size_t extensionSize;
    error = clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, 0, NULL, &extensionSize);
    char* extensions = (char*) malloc(sizeof(char) * extensionSize+1);
    error |= clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, extensionSize, extensions, NULL);
    assert(error == CL_SUCCESS);
    extensions[extensionSize] = '\0';
    if(strstr(extensions, "cl_khr_gl_sharing") == 0)
    {
        printf("OpenCL-OpenGL Interoperability not supported\n");
        exit(1);
    }    
    
    // Get Context from OpenGL
    #if defined (__APPLE__) || defined(MACOSX)
    CGLContextObj kCGLContext = CGLGetCurrentContext();
    CGLShareGroupObj kCGLShareGroup = CGLGetShareGroup(kCGLContext);
    cl_context_properties properties[] = {
        CL_CONTEXT_PROPERTY_USE_CGL_SHAREGROUP_APPLE, 
        (cl_context_properties)kCGLShareGroup,
        0};
    #elif defined(WIN32)
    cl_context_properties properties[] = {
        CL_GL_CONTEXT_KHR, 
        (cl_context_properties) wglGetCurrentContext(),
        CL_WGL_HDC_KHR,
        (cl_context_properties) wglGetCurrentDC(),
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties) platform,
        0};
    #else
    cl_context_properties properties[] = {
       CL_GL_CONTEXT_KHR, (cl_context_properties) glXGetCurrentContext(),
       CL_GLX_DISPLAY_KHR, (cl_context_properties) glXGetCurrentDisplay(),
       CL_CONTEXT_PLATFORM, (cl_context_properties) platform,
       0};    
    #endif
    
    context = clCreateContext(properties, 1, &device, NULL, NULL, &error);
    assert (error == CL_SUCCESS);
    
    //Queue
    queue = clCreateCommandQueue(context, device, 0, &error);
    assert (error == CL_SUCCESS);
         
    ///Load the source code containing the kernel
    char* sourceStr = loadTextFile(KER_DIR"tetra_kernal.cl");
        
    // Creates the program
    program = clCreateProgramWithSource(context, 1, (const char **)&sourceStr, NULL, &error);
    assert(error == CL_SUCCESS);
    free(sourceStr);

    // Builds the program
    error = clBuildProgram(program, 1, &device, NULL, NULL, NULL);   
    // Shows the log
    char* buildLog;
    size_t logSize;
    // First call to know the proper size
    clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, 0, NULL, &logSize);
    buildLog = malloc(sizeof(char)*(logSize+1));
    // Second call to get the log
    clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, logSize, buildLog, NULL);
    buildLog[logSize] = '\0';
    if(error != CL_SUCCESS)
    {
        printf("Build Log:\n");
        printf("==========\n"); 
        puts(buildLog);
        free(buildLog);
        printf("-----------------\n");
        exit(1);
    }    
    
    // Check maximum workgroup size
    size_t maxWorkgroupSize = 0;
    error = clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(size_t), &maxWorkgroupSize, NULL);
    assert(error == CL_SUCCESS);
    workgroupSize = min(maxWorkgroupSize, workgroupSize);       
    
    // Extracting kernals   
    fillMpIndex_kernel = clCreateKernel(program, "fillMpIndex", &error);
    assert(error == CL_SUCCESS); 
        
    
    enumActiveMPs_kernel = clCreateKernel(program, "enumActiveMPs", &error);
    assert(error == CL_SUCCESS);
    
    calcVoxNormals_kernel =  clCreateKernel(program, "calcVoxNormals", &error);
    assert(error == CL_SUCCESS);    
    
    calcVertsAndNorms_kernel = clCreateKernel(program, "calcVertsAndNorms", &error);
    assert(error == CL_SUCCESS);

      
    calcTriangs_kernel = clCreateKernel(program, "calcTriangs", &error);
    assert(error == CL_SUCCESS);  
    
    // Initialize Scan Kernels            
    scanner = oclCreateScanner(context, queue, device, workgroupSize, maxWorkItemsCount,KER_DIR"defaultScanner.cl", sizeof(int), 0, NULL, &error);
    assert(error == CL_SUCCESS);
    
    activeScanner = oclCreateScanner(context, queue, device, workgroupSize, maxWorkItemsCount,KER_DIR"activeScanner.cl", sizeof(int),  0, NULL, &error);
    assert(error == CL_SUCCESS);
    
    vertsScanner = oclCreateScanner(context, queue, device, workgroupSize, maxWorkItemsCount,KER_DIR"vertsScanner.cl", sizeof(int), 1, vertsScanParaSizes,&error);
    assert(error == CL_SUCCESS);    
    
    triangsScanner = oclCreateScanner(context, queue, device, workgroupSize, maxWorkItemsCount,KER_DIR"triangsScanner.cl", sizeof(int), 2, triangsScanParaSizes ,&error);
    assert(error == CL_SUCCESS);        
    
    clReleaseProgram(program);
    // Load field and compute normals per voxels using finite differences
    loadField(fieldFileName);
    
    printf("Calculating Voxel Normals (once and for all)...\n");
    calcVoxNormals();
    
    
    //Allocate Fixed-Sized Memory    
    mpIndex = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(unsigned char)*(totalMPs), 0, &error);
    assert (error == CL_SUCCESS);
    
    mpActiveId = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(int)*(totalMPs), 0, &error);
    assert (error == CL_SUCCESS);      
}

void oclTetraClose()
{    
    //Release Scan Kernels
    oclReleaseScanner(scanner);
    oclReleaseScanner(activeScanner);
    oclReleaseScanner(vertsScanner);
    oclReleaseScanner(triangsScanner);
    //Memory
    clReleaseMemObject(field);   
    clReleaseMemObject(voxNormals);
    clReleaseMemObject(mpIndex);
    clReleaseMemObject(mpActiveId);
    
    //Kernals    
    clReleaseKernel(fillMpIndex_kernel);
    clReleaseKernel(enumActiveMPs_kernel);
    clReleaseKernel(calcVoxNormals_kernel);
    clReleaseKernel(calcVertsAndNorms_kernel);
    clReleaseKernel(calcTriangs_kernel);
    
    //Others
    
    clReleaseCommandQueue(queue);
    clReleaseContext(context);
}
